import { useState } from 'react';
import { 
  WagmiConfig, createClient, useAccount, 
  useConnect, useDisconnect, useContractReads, 
  configureChains, useContractWrite, usePrepareContractWrite,
  useWaitForTransaction, useContractRead
} from 'wagmi';
import { mainnet } from 'wagmi/chains';
import { publicProvider } from 'wagmi/providers/public';
import { infuraProvider } from 'wagmi/providers/infura';
import { InjectedConnector } from 'wagmi/connectors/injected';
import { readContract, writeContract, prepareWriteContract, waitForTransaction } from '@wagmi/core'
import { ethers, BigNumber } from 'ethers';
import Noty from 'noty';

import ABI from './abi.json';
import Twitter from './img/twitter.png';
import Looksrare from './img/looksrare.png';
import Opensea from './img/opensea.png';
import X2Y2 from './img/x2y2.png';
import 'bulma/css/bulma.min.css';
import 'noty/lib/noty.css';  
import 'noty/lib/themes/relax.css';  

// 0x5FbDB2315678afecb367f032d93F642f64180aa3 - local
// 0xd8977312564d808e8e57778acd42c16f70d1cd41 - goerli
// 0x25982b72416BBF379e6A28C694baD6559E757c11 - mainnet
const contractAddress = "0x25982b72416BBF379e6A28C694baD6559E757c11";

const { chains, provider } = configureChains(
  [mainnet],
  [infuraProvider({ apiKey: 'e7674eea9d72437d9129b41ed018183d' }), publicProvider()]
);

const connector = new InjectedConnector({ chains });

const client = createClient({
  autoConnect: true,
  connectors: [connector],
  provider
});

function Contract() {
  const { address, isConnected } = useAccount();
  const { connect } = useConnect({connector});
  const { disconnect } = useDisconnect();
  const _contract = {address: contractAddress, abi: ABI}
  const [isLoading, setIsLoading] = useState(0);
  const [options, setOptions] = useState({ 
    totalSupply: null,
    max: null,
    price: 0,
    winners: 0,
    winningAmount: 0,
    phraseSet: false,
    mintAmount: 1,
    balance: 0,
    hash: '',
    txPending: false,
    tokensSet: '',
    winningTokens: '',
    claimableTokens: ''
  });
  function handleStateChange(obj) {
    setOptions(preState => ({...preState , ...obj}));
  }
  useContractReads({
    contracts: [
      {..._contract, functionName: 'totalSupply'},
      {..._contract, functionName: 'max'},
      {..._contract, functionName: 'price'},
      {..._contract, functionName: 'winners'},
      {..._contract, functionName: 'winningAmount'},
      {..._contract, functionName: 'phraseSet'},
      {..._contract, functionName: 'balanceOf', args: [address]},
      {..._contract, functionName: 'winningPhraseHash'},
      {..._contract, functionName: 'tokensSet'}
    ],
    watch: true,
    cacheTime: 4_000,
    onSuccess(data) {
      handleStateChange({
          totalSupply: Number(data[0]),
          max: Number(data[1]),
          price: Number(data[2]),
          winners: Number(data[3]),
          winningAmount: Number(data[4]),
          phraseSet: data[5],
          balance: Number(data[6]),
          hash: data[7],
          tokensSet: data[8]
      });
    }
  });
  useContractRead({
    ..._contract,
    enabled: options.totalSupply > 0 && options.totalSupply === options.max && options.tokensSet,
    functionName: 'getWinningTokens',
    watch: true,
    cacheTime: 8000,
    onSuccess(data) {
      handleStateChange({
        winningTokens: JSON.stringify(data)
      });
    }
  });
  const mintPrepare = usePrepareContractWrite({
    address: contractAddress,
    abi: ABI,
    enabled: isConnected && options.mintAmount > 0 && options.totalSupply < options.max,
    functionName: 'mint',
    args: [options.mintAmount],
    overrides: {
        from: address,
        value: BigNumber.from((options.price * options.mintAmount).toString())
    }
  });
  const mintWrite = useContractWrite({
    ...mintPrepare.config,
    onError(data) {
      new Noty({
        type: 'error',
        text: `Tx rejected`,
        theme: 'relax',
        layout: 'topCenter',
        timeout: 3000
      }).show();
      mintWrite.reset();
      handleStateChange({txPending: false});
    },
    onSuccess(data) {
      new Noty({
        type: 'info',
        text: `Tx ${data.hash} sent...pending confirmation`,
        theme: 'relax',
        layout: 'topCenter',
        timeout: 3000
      }).show();
    }
  });
  useWaitForTransaction({
    hash: mintWrite.data?.hash,
    enabled: mintWrite.status === 'success',
    onSuccess(data) {
      handleStateChange({txPending: false});
      new Noty({
        type: 'success',
        text: `Tx ${data.transactionHash} confirmed`,
        theme: 'relax',
        layout: 'topCenter',
        timeout: 3000
      }).show();
    }
  });
  async function checkClaimable() {
    const tokens = JSON.parse(options.winningTokens);
    let claimable = [];
    for (let i = 0; i < tokens.length; i++) {
      const tokenId = BigNumber.from(tokens[i]).toString();
      const ownerOf = await readContract({
        address: contractAddress,
        abi: ABI,
        functionName: 'ownerOf',
        args: [tokenId]
      });
      const isWinner = await readContract({
        address: contractAddress,
        abi: ABI,
        functionName: 'winningTokens',
        args: [tokenId]
      });
      const isClaimed = await readContract({
        address: contractAddress,
        abi: ABI,
        functionName: 'tokenClaimed',
        args: [tokenId]
      });
      if (
        ownerOf === address &&
        isWinner === true &&
        isClaimed === false
      ) {
        claimable.push(tokenId);
      }
    }
    handleStateChange({claimableTokens: JSON.stringify(claimable)});
    if (claimable.length > 0) {
      new Noty({
        type: 'success',
        text: `Winner! Congratulations, you have ${claimable.length} tokens elligible to claim the Ether prize for.`,
        theme: 'relax',
        layout: 'topCenter',
        timeout: 8000
      }).show();
    } else {
      new Noty({
        type: 'error',
        text: `Too bad! None of your tokens are elligible to claim the Ether prize.`,
        theme: 'relax',
        layout: 'topCenter',
        timeout: 8000
      }).show();
    }
  }
  async function claimToken(tokenId) {
    const config = await prepareWriteContract({
      ..._contract,
      functionName: 'claimToken',
      args: [tokenId]
    });
    const { hash } = await writeContract(config);
    new Noty({
      type: 'info',
      text: `Claim for token ${tokenId} has been sent in Tx ${hash}.`,
      theme: 'relax',
      layout: 'topCenter',
      timeout: 8000
    }).show();
    const receipt = await waitForTransaction({
      hash,
      timeout: 90000
    });
    if (receipt) {
      new Noty({
        type: 'success',
        text: `Claim for token ${tokenId} has succeeded!`,
        theme: 'relax',
        layout: 'topCenter',
        timeout: 8000
      }).show();
    }
  }
  if (isConnected && options.phraseSet)
    return (
      <>
        <button className="button" style={{marginBottom: '1em'}} onClick={() => disconnect()}>Disconnect Wallet</button>
        {options.totalSupply === options.max && options.tokensSet === true && (
          <button className={"button is-info " + (isLoading === 1 && "is-loading")} style={{marginBottom: '1em', marginLeft: '1em'}} onClick={async () => {
            setIsLoading(1);
            await checkClaimable().then(() => setIsLoading(0));
          }}>Check Claimable Tokens</button>
        )}
        {options.totalSupply === options.max && options.tokensSet === false && (
          <button className="button is-warning" style={{marginBottom: '1em', marginLeft: '1em'}} disabled>Tokens not entered yet. Go degen.</button>
        )}

        {options.claimableTokens !== '' && (
          <div className="buttons">
            {JSON.parse(options.claimableTokens).map((tokenId, idx) => (
              <button key={idx} className={"button is-success " + (isLoading === 1 && "is-loading")} style={{marginBottom: '1em', marginRight: '1em'}} onClick={async () => {
                setIsLoading(1);
                await claimToken(tokenId).then(async d => {
                  await checkClaimable().then(() => setIsLoading(0));
                }).catch(() => setIsLoading(0));
              }}>
                Claim Token {tokenId}
              </button>
            ))}
          </div>
        )}

        <div className="block">
          <div className="buttons has-addons">
            <a href="https://twitter.com/intent/tweet?text=Just minted a lottery ticket to win 1 ETH at https://lzahq.tech/afouagmi by @lza_menace! LFG!" target={"_blank"} alt="twitter share" className="button"><img src={Twitter} width={"32px"} /></a>
            <a href="#" target={"_blank"} alt="looksrare link" className="button"><img src={Looksrare} width={"32px"} /></a>
            <a href="#" target={"_blank"} alt="x2y2 link" className="button"><img src={X2Y2} width={"32px"} /></a>
            <a href="#" target={"_blank"} alt="opensea link" className="button"><img src={Opensea} width={"32px"} /></a>
          </div>
        </div>
      
        <hr></hr>

        {options.totalSupply < options.max && (
          <div className="block">
            <h1 className="heading">Price: {ethers.utils.formatEther((options.price * options.mintAmount).toString())} ETH</h1>
            <div className="field has-addons">
              <div className="control">
                <input className="input" type="text" placeholder="Mint amount" value={options.mintAmount} onChange={(v) => {
                  if (v.target.value >= 0) handleStateChange({mintAmount: Number(v.target.value)})}
                }></input>
              </div>
              <div className="control">
                <button className={"button is-info " + ((options.txPending) && "is-loading")} disabled={mintPrepare.status === 'error'} onClick={() => {
                  mintWrite.write?.();
                  handleStateChange({txPending: true});
                }}>
                  Mint
                </button>
              </div>
            </div>
            <h1 className="heading"><strong>You'll have a ~{((options.balance + options.mintAmount) / options.max * 100).toFixed(2)}% chance to win 1 ETH</strong></h1>
          </div>
        )}
        <div className="block">
          <h1 className="heading">Connected to {address.slice(0,6) + '...' + address.slice(-4)} :: <strong>You own {options.balance} tokens</strong></h1>
          {options.totalSupply !== options.max && (
            <>
              <br />
              <h1 className="heading">Minted {options.totalSupply} / {options.max} {options.max - options.totalSupply > 0 && (
                <>
                  <strong>({options.max - options.totalSupply} remaining)</strong>
                </>
              )}</h1>
              <progress className="progress is-primary" value={options.totalSupply} max={options.max}></progress>
            </>
          )}
        </div>
        {options.tokensSet && (
          <div className="notification is-warning">
            Tokens have now been specified in the contract. Owners of those tokens can claim the Ether by clicking the claim buttons above. Congrats to the winners!
          </div>
        )}
        {options.totalSupply === options.max && (
          <div className="notification is-success">
            All lottery tokens have been sold! Ether claims will be made available to winners approximately 48 hours after mint out.
          </div>
        )}
        <div className="notification is-info">
          The secret hash containing the pre-determined winners is: <strong>{options.hash}</strong>
        </div>
      </>
    )
  if (isConnected && !options.phraseSet)
    return (
      <>
        <h3 style={{fontSize: '2em'}}>Not minting yet. Come back later.</h3>
      </>
    )
  return <button className="button is-dark" onClick={() => connect()}>Connect Wallet</button>
}

function App() {
  return (
    <div className="container">
      <WagmiConfig client={client}>
        <section className="hero is-medium">
          <div className="hero-body">
            <p className="title">
              A Few Of Us Are Gonna Make It.
            </p>
            <p className="subtitle">
              An NFT Lottery. 10 tokens are pre-determined to win.<br />
              Mint one for your chance to win 1 ETH!
            </p>
            <Contract />
          </div>
        </section>
      </WagmiConfig>
    </div>
  );
}

export default App;
